//-----------------------------------------------------------------------------
// Title      : AsyncArt micropipeline demo
// Project    : AsyncArt
// URL        : https://www.ohwr.org/project/asyncart
//-----------------------------------------------------------------------------
// File       : asyncart_demo.v
// Author(s)  : Javier Garcia Lasheras <jgarcia@gl-research.com>
// Company    : GL Research Corporation SLU
//-----------------------------------------------------------------------------
// Description: Demonstrative Verilog demo for variable depth micropipelines.
//-----------------------------------------------------------------------------
// Copyright (c) 2006-2019 GL Research Corporation SLU
//-----------------------------------------------------------------------------
// GNU LESSER GENERAL PUBLIC LICENSE
//-----------------------------------------------------------------------------
// This source file is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published by the
// Free Software Foundation; either version 2.1 of the License, or (at your
// option) any later version. This source is distributed in the hope that it
// will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details. You should have
// received a copy of the GNU Lesser General Public License along with this
// source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
//-----------------------------------------------------------------------------

`include "asyncart_source.v"
`include "asyncart_sink.v"
`include "asyncart_reg.v"

module sync_register_32b (
    output [31:0] data_out,
    input [31:0] data_in,
    input clk,
    input enable,
    input reset);
    reg [31:0] data_reg;
    always @(posedge clk)
    if (reset) begin
        data_reg <= 32'b0 ;
    end else if (enable) begin
        data_reg <= data_in;
    end
    assign data_out = data_reg;
endmodule


module asyncart_demo(
	input act_in,
	input rst_in,
	output [31:0] data_out,
	);


	wire reset = 0;
	wire enable = 1;

        parameter micropipeline_depth = 3;

        wire [micropipeline_depth-1:0] fire;
        wire [micropipeline_depth-1:0] phase;
        wire [31:0] data_link [micropipeline_depth:0];
        wire act, rst;

        assign act = act_in;
        assign rst = rst_in;


        asyncart_source asyncart_source_instance (
            .fire_out(fire[0]),
            .phase_out(phase[0]),
            .phase_in_neg(phase[1]),
            .act(act),
            .rst(rst)
        );


        genvar i;
        generate
            for (i=1; i<micropipeline_depth-1; i=i+1) begin : asyncart_reg_block
            asyncart_reg asyncart_reg_instance (
                .fire_out(fire[i]),
                .phase_out(phase[i]),
                .phase_in_pos(phase[i-1]),
                .phase_in_neg(phase[i+1]),
                .act(act),
                .rst(rst)
            );
        end 
        endgenerate


        asyncart_sink asyncart_sink_instance (
            .fire_out(fire[micropipeline_depth-1]),
            .phase_out(phase[micropipeline_depth-1]),
            .phase_in_pos(phase[micropipeline_depth-2]),
            .act(act),
            .rst(rst)
        );



        genvar j;
        generate
            for (j=0; j<=micropipeline_depth-1; j=j+1) begin : sync_register_block
            sync_register_32b sync_register_32b_instance (
                .data_in(data_link[j]),
                .data_out(data_link[j+1]),
                .clk(fire[j]),
                .enable(enable),
                .reset(reset)
            );
        end
        endgenerate


        // Implement an Up-Counter in the source register (first pipeline stage)
	assign data_link[0] = data_link[1] + 1;

        // Connect to the output the contents of the sink register  (last pipeline stage)
	assign data_out = data_link[micropipeline_depth];

endmodule
